/*----------------------------------------------------------------------
  File    : gramod1.c
  Contents: graphical model management (based on directed graphs)
            base functions and (optional) parsing
  Author  : Christian Borgelt
  History : 1996.02.23 file created as ascons.c
            1996.02.24 functions debugged
            1996.03.17 adapted to attribute types
            1996.03.26 adapted to modified ptree functions
            1996.04.09 functions gm_rand() and order() modified
            1996.04.12 deletion of candidate vector in gm_delete() added
            1996.06.18 output of frequencies changed to relative numbers
            1996.06.23 line length checks debugged
            1996.07.26 output precision reduced
            1996.11.15 description modes added
            1996.11.27 adapted to floating point counters
            1997.02.28 adapted to prob./poss. tree types
            1997.03.01 function gm_exec() added, redesign
            1997.03.10 bug in function gm_exec() fixed
            1997.08.28 bug in functions gm_aggr() fixed
            1998.02.08 function gm_parse() transferred from parse.c
            1998.02.14 description and parsing of links added
            1998.02.20 zero frequency output suppression removed
            1998.06.23 adapted to modified attset functions
            1999.03.01 adapted to extended prob./poss. tree structure
            1999.04.05 parameters 'attid', 'conid' added to gm_check()
            1999.04.30 parameter 'all' added to function gm_parse()
            1999.10.22 missing init. of GMATT->deps added to gm_parse()
            1999.12.02 parameter 'corr' added to function gm_exec()
            2001.05.26 adapted to changed module scan
            2002.02.08 adapted to changed module ptree
            2002.03.03 module renamed to gramod
            2002.03.11 creation of default model added to gm_parse
            2002.07.16 parameter 'delnodes' added to function gm_clear()
            2006.02.16 restriction of number of conditions removed
            2007.01.29 top. order no longer required for poss. models
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "gramod.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE   32            /* block size for candidate vector */

/* --- error messages --- */
/* error codes 0 to -15 defined in scanner.h */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_UNKATT    (-17)       /* unknown attribute */
#define E_DUPATT    (-18)       /* duplicate attribute */
#define E_MISATT    (-19)       /* missing attribute */
#define E_DUPCDL    (-20)       /* duplicate candidate list */
#define E_RANGE     (-21)       /* invalid candidate range */
#define E_CAND      (-22)       /* invalid candidate */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef GM_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_UNKATT  -17 */  "#unknown attribute '%s'",
  /* E_DUPATT  -18 */  "#duplicate attribute '%s'",
  /* E_MISATT  -19 */  "#missing attribute '%s'",
  /* E_DUPCDL  -20 */  "#duplicate candidate list for attribute '%s'",
  /* E_RANGE   -21 */  "#invalid candidate range",
  /* E_CAND    -22 */  "#invalid candidate '%s'",
};
#endif

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static int order (GRAMOD *gm)
{                               /* --- find a topological order */
  ATTID i, k, n, prv;           /* loop variables, counters */
  GMATT *att;                   /* to traverse the attributes */
  ATTID *ord;                   /* to traverse the topological order */

  assert(gm);                   /* check the function argument */
  ord = gm->order;              /* get the topological order vector */
  att = gm->atts;               /* and traverse the attributes */
  for (i = n = 0; i < gm->attcnt; i++, att++) {
    if (att->ptree              /* if there are conditions */
    && (pt_concnt(att->ptree) > 0)) {
      att->pos = -1; continue;} /* clear the attribute position, */
    *ord++   = i;               /* otherwise add the attribute */
    att->pos = n++;             /* to the topological order */
  }                             /* (collect parentless attributes) */
  while (n < gm->attcnt) {      /* while there are attributes left */
    prv = n;                    /* note the old counter value */
    att = gm->atts;             /* and traverse the attributes */
    for (i = 0; i < gm->attcnt; i++, att++) {
      if (att->pos >= 0)        /* skip all attributes that are */
        continue;               /* already in the topological order */
      for (k = pt_concnt(att->ptree); --k >= 0; )
        if (gm->atts[pt_conid(att->ptree, k)].pos < 0)
          break;                /* skip attributes with a condition */
      if (k >= 0) continue;     /* that is not yet in the top. order */
      *ord++   = i;             /* otherwise add the attribute */
      att->pos = n++;           /* to the topological order */
    }
    if (n <= prv) {             /* if no attribute was added, */
      gm->order[0] = -1; return -1; } /* there is at least one loop, */
  }                             /* so invalidate the order and abort */
  return 0;                     /* return 'ok' */
}  /* order() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static GRAMOD* create (ATTSET *attset, int type)
{                               /* --- create a graphical model */
  ATTID  i;                     /* number of atts. and loop variable */
  GRAMOD *gm;                   /* created graphical model */
  GMATT  *att;                  /* to traverse the attributes */

  assert(attset);               /* check the function argument */
  i  = as_attcnt(attset);       /* get the number of attributes */
  gm = (GRAMOD*)calloc(1, sizeof(GRAMOD) +(size_t)(i-1) *sizeof(GMATT));
  if (!gm) return NULL;         /* allocate the base structure */
  gm->order    = (ATTID*)malloc(2 *(size_t)i *sizeof(ATTID));
  if (!gm->order) { free(gm); return NULL; }
  gm->buf    = gm->order +i;    /* create and organize vectors */
  gm->type   = type & GM_POSS;  /* set the tree type (prob./poss.) */
  gm->attset = attset;          /* note the attribute set */
  gm->attcnt = i;               /* and the number of attributes */
  gm->concnt = 0;               /* clear the number of conditions */
  gm->tplcnt = 0;               /* the number of tuples, */
  gm->eval   = 0;               /* and the evaluation result */
  gm->order[0] = -1;            /* invalidate the topological order */
  for (att = gm->atts +i; --i >= 0; )  /* note the */
    (--att)->att = as_att(attset, i);  /* corresp. attributes */
  return gm;                    /* return created graphical model */
}  /* create() */

/*--------------------------------------------------------------------*/

GRAMOD* gm_create (ATTSET *attset, int type)
{                               /* --- create a graphical model */
  ATTID  i;                     /* loop variable */
  GRAMOD *gm;                   /* created graphical model */
  GMATT  *att;                  /* to traverse the attributes */

  assert(attset);               /* check the function argument */
  gm = create(attset, type);    /* create a graphical model */
  if (!gm) return NULL;         /* and traverse the attributes */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
    if (att_type((--att)->att) != AT_NOM)
      continue;                 /* skip non-nominal attributes */
    att->ptree = pt_create(att->att, type);
    if (!att->ptree) { gm_delete(gm, 0); return NULL; }
  }                             /* create prob./poss. trees */
  return gm;                    /* return created graphical model */
}  /* gm_create() */

/*--------------------------------------------------------------------*/

void gm_delete (GRAMOD *gm, int delas)
{                               /* --- delete a graphical model */
  ATTID i;                      /* loop variable */

  assert(gm);                   /* check the function argument */
  for (i = 0; i < gm->attcnt; i++) {
    if (gm->atts[i].ptree) pt_delete(gm->atts[i].ptree, 0);
    if (gm->atts[i].cdds) free(gm->atts[i].cdds);
    if (gm->atts[i].deps) free(gm->atts[i].deps);
  }                             /* delete prob./poss tree and vectors */
  if (delas) as_delete(gm->attset);
  free(gm->order);              /* delete attribute set, top. order, */
  free(gm);                     /* and graphical model body */
}  /* gm_delete() */

/*--------------------------------------------------------------------*/

double gm_setnorm (GRAMOD *gm, double tplcnt)
{                               /* --- set number of tuples for norm. */
  ATTID i;                      /* loop variable */

  assert(gm && (tplcnt >= 0));  /* check the function arguments */
  for (i = 0; i < gm->attcnt; i++)
    if (gm->atts[i].ptree) pt_setnorm(gm->atts[i].ptree, tplcnt);
  return gm->tplcnt = tplcnt;   /* update the prob./poss. trees */
}  /* gm_setnorm() */           /* and note the number of tuples */

/*--------------------------------------------------------------------*/

float gm_setlc (GRAMOD *gm, float lcorr)
{                               /* --- set Laplace correction */
  ATTID i;                      /* loop variable */

  assert(gm && (lcorr >= 0));   /* check the function arguments */
  for (i = 0; i < gm->attcnt; i++)
    if (gm->atts[i].ptree) pt_setlc(gm->atts[i].ptree, lcorr);
  return gm->lcorr = lcorr;     /* update the prob./poss. trees */
}  /* gm_setlc() */             /* and note the Laplace correction */

/*--------------------------------------------------------------------*/

ATTID gm_consum (GRAMOD *gm)
{                               /* --- get the number of conitions */
  ATTID i, n = 0;               /* loop variable, counter */

  assert(gm);                   /* check the function argument */
  for (i = 0; i < gm->attcnt; i++)
    if (gm->atts[i].ptree) n += pt_concnt(gm->atts[i].ptree);
  return n;                     /* sum the number of conditions */
}  /* gm_consum() */            /* of the prob./poss. trees */

/*--------------------------------------------------------------------*/

ATTID gm_parsum (GRAMOD *gm)
{                               /* --- get the number of parameters */
  ATTID i, n = 0;               /* loop variable, counter */

  assert(gm);                   /* check the function argument */
  for (i = 0; i < gm->attcnt; i++)
    if (gm->atts[i].ptree) n += pt_parcnt(gm->atts[i].ptree);
  return n;                     /* sum the number of parameters */
}  /* gm_parsum() */            /* of the prob./poss. trees */

/*--------------------------------------------------------------------*/

ATTID gm_cddadd (GRAMOD *gm, ATTID attid, ATTID cddid)
{                               /* --- add a condition candidate */
  GMATT *att;                   /* to access the attribute */
  ATTID size;                   /* new candidate vector size */
  ATTID *vec;                   /* new candidate vector */

  assert(gm && (cddid != attid) /* check the function arguments */
  &&    (attid >= 0) && (attid < gm->attcnt)
  &&    (cddid >= 0) && (cddid < gm->attcnt));
  att = gm->atts +attid;        /* get the attribute */
  #ifndef NDEBUG                /* if to compile a debug version */
  for (size = att->cddcnt; --size >= 0; )
    assert(att->cdds[size] != cddid);
  #endif                        /* check for a duplicate candidate */
  if (att->cddcnt >= att->cddsize) {
    size  = att->cddsize;       /* if the candidate vector is full */
    size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
    vec   = (ATTID*)realloc(att->cdds, (size_t)size *sizeof(ATTID));
    if (!vec) return -1;        /* allocate a new candidate vector */
    att->cdds = vec; att->cddsize = size;
  }                             /* set the new vector and its size */
  att->cdds[att->cddcnt] = cddid;  /* add the new candidate */
  return att->cddcnt++;            /* and return its index */
}  /* gm_cddadd() */

/*--------------------------------------------------------------------*/

void gm_cddrem (GRAMOD *gm, ATTID attid, ATTID index)
{                               /* --- remove a condition candidate */
  GMATT *att;                   /* to access the attribute */
  ATTID *p;                     /* to traverse the candidate vector */

  assert(gm                     /* check the function arguments */
  &&    (attid >= 0) && (attid < gm->attcnt));
  att = gm->atts +attid;        /* get the attribute */
  assert(index < att->cddcnt);  /* check the function arguments */
  if (index < 0)                /* if to remove all candidates, */
    att->cddcnt = 0;            /* clear the candidate counter */
  else {                        /* if to remove one candidate, */
    p = att->cdds +index;       /* traverse the candidate vector */
    for (index = --att->cddcnt -index; --index >= 0; ) {
      p[0] = p[1]; p++; }       /* shift left/up all candidates */
  }                             /* following the one to remove */
}  /* gm_cddrem() */

/*--------------------------------------------------------------------*/

ATTID gm_conadd (GRAMOD *gm, ATTID attid, ATTID conid)
{                               /* --- add a condition */
  ATTID cnt;                    /* number of conditions */
  GMATT *att;                   /* to access the attribute */
  GMATT *con;                   /* to access the condition */
  ATTID size;                   /* new dependent atts. vector size */
  ATTID *vec;                   /* new dependent atts. vector */

  assert(gm && (conid != attid) /* check the function arguments */
  &&    (attid >= 0) && (attid < gm->attcnt)
  &&    (conid >= 0) && (conid < gm->attcnt)
  &&    (att_type(gm->atts[attid].att) == AT_NOM)
  &&    (att_type(gm->atts[conid].att) == AT_NOM));
  att = gm->atts +attid;        /* get the attribute */
  con = gm->atts +conid;        /* and the condition */
  cnt = pt_concnt(att->ptree);  /* get number of existing conditions */
  #ifndef NDEBUG                /* if to compile a debug version, */
  for (size = cnt; --size >= 0;)/* check for a duplicate condition */
    assert(pt_conid(att->ptree, size) != conid);
  #endif
  if (con->depcnt >= con->depsize) {
    size  = con->depsize;       /* if dependent atts. vector is full */
    size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
    vec   = (ATTID*)realloc(con->deps, (size_t)size *sizeof(ATTID));
    if (!vec) return -1;        /* (re)allocate a dep. atts. vector */
    con->deps = vec; con->depsize = size;
  }                             /* set the new vector and its size */
  if (pt_conadd(att->ptree, cnt, gm->atts[conid].att) < 0)
    return -1;                  /* add the condition to the tree */
  con->deps[con->depcnt++] = attid; /* add a dependent attribute */
  if (con->pos > att->pos)      /* check positions in top. order and */
    gm->order[0] = -1;          /* invalidate the order if necessary */
  gm->concnt++;                 /* increment the condition counter */
  return cnt;                   /* return the index of the condition */
}  /* gm_conadd() */

/*--------------------------------------------------------------------*/

void gm_conrem (GRAMOD *gm, ATTID attid, ATTID index)
{                               /* --- remove a condition */
  ATTID i, k, n;                /* loop variables, buffer */
  GMATT *att;                   /* to access the attribute/condition */
  PTREE *pt;                    /* probability/possibility tree */

  assert(gm && (attid < gm->attcnt)  /* check the function arguments */
  &&    ((attid < 0) || (att_type(gm->atts[attid].att) == AT_NOM)));
  if (attid < 0) {              /* if to remove all conditions */
    for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
      if (!(--att)->ptree) continue;
      pt_conrem(att->ptree, -1, 0);
      att->depcnt = 0;          /* traverse the attributes */
    }                           /* and remove all conditions */
    gm->concnt = 0; return;     /* reinit. the condition counter */
  }                             /* and abort the function */
  pt = gm->atts[attid].ptree;   /* get the prob./poss. tree and */
  assert(index < pt_concnt(pt));/* check the condition index */
  if (index < 0) { n = 0;     i = pt_concnt(pt); }
  else           { n = index; i = index +1; }
  gm->concnt -= i-n;            /* decrement the condition counter */
  while (--i >= n) {            /* traverse the conditions to remove */
    att = gm->atts +pt_conid(pt, i);
    for (k = att->depcnt; --k >= 0; ) {
      if (att->deps[k] == attid) {
        att->deps[k] = att->deps[--att->depcnt]; break; }
    }                           /* traverse dependent atts. vector */
  }                             /* and remove the attribute from it */
  pt_conrem(pt, index, 0);      /* remove condition(s) from the tree */
  /* A topological order of the attributes (gm->order), which may */
  /* exist, remains valid, because the new set of conditions is */
  /* weaker than the old (there are fewer edges in the graph). */
}  /* gm_conrem() */

/*--------------------------------------------------------------------*/

void gm_clear (GRAMOD *gm, int delnodes)
{                               /* --- clear distributions */
  ATTID i;                      /* loop variable */

  assert(gm);                   /* check the function argument */
  for (i = 0; i < gm->attcnt; i++)
    if (gm->atts[i].ptree) pt_clear(gm->atts[i].ptree, delnodes);
}  /* gm_clear() */             /* clear the prob./poss. tree(s) */

/*--------------------------------------------------------------------*/

int gm_aggr (GRAMOD *gm, int distuv)
{                               /* --- aggregate instantiation weight */
  ATTID  i;                     /* loop variable */
  GMATT  *att;                  /* to traverse the attributes */
  WEIGHT wgt;                   /* weight of the instantiation */

  assert(gm);                   /* check the function arguments */
  wgt = as_getwgt(gm->attset);  /* get the instantiation weight */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; )
    if ((--att)->ptree && (pt_aggr(att->ptree, wgt, distuv) < 0))
      return -1;                /* aggregate the instantiation weight */
  return 0;                     /* return 'ok' */
}  /* gm_aggr() */

/*--------------------------------------------------------------------*/

static int check (GMATT *atts, ATTID attid, ATTID conid)
{                               /* --- recursively check for a loop */
  ATTID i;                      /* loop variable */
  ATTID *dep;                   /* to traverse the dep. att. ids. */

  assert(atts                   /* check the function arguments */
  &&    (attid >= 0) && (conid >= 0));
  if (atts[attid].pos > atts[conid].pos)
    return 0;                   /* check positions in top. order */
  for (dep = atts[attid].deps +(i = atts[attid].depcnt); --i >= 0; )
    if ((*--dep == conid) || (check(atts, *dep, conid) != 0))
      return -1;                /* if the condition is among the */
  return 0;                     /* descendent attributes, abort, */
}  /* check() */                /* otherwise return 'ok' */

/*--------------------------------------------------------------------*/

int gm_check (GRAMOD *gm, ATTID attid, ATTID conid)
{                               /* --- check graphical model */
  assert(gm);                   /* check the function argument */
  if (gm->type == GM_POSS)      /* for a possibilistic model, */
    return 0;                   /* no check is needed */
  if ((gm->order[0] < 0)        /* if no topological order exists */
  &&  (order(gm) != 0))         /* and it cannot be created either, */
    return -1;                  /* abort the function with failure */
  if (attid < 0) return 0;      /* if no edge to test is given, abort */
  assert((attid < gm->attcnt)   /* check the function arguments */
  &&     (conid < gm->attcnt) && (conid >= 0));
  if (attid == conid)           /* if the attributes are identical, */
    return -1;                  /* there is a direct loop */
  return check(gm->atts, attid, conid);
}  /* gm_check() */             /* recursively check for a loop */

/*--------------------------------------------------------------------*/

int gm_rand (GRAMOD *gm, RANDFN randfn)
{                               /* --- compute random instances */
  ATTID i;                      /* loop variable */
  ATTID *o;                     /* to traverse the topological order */
  GMATT *att;                   /* to traverse the attributes */

  assert(gm && randfn           /* check the function arguments */
  &&    (gm->type == GM_PROB)); /* (works only for prob. models) */
  if ((gm->order[0] < 0)        /* if no topological order exists */
  &&  (order(gm) != 0))         /* and it cannot be created either, */
    return -1;                  /* abort the function */
  o = gm->order;                /* traverse in topological order */
  for (i = gm->attcnt; --i >= 0; ) {
    att = gm->atts +*o++;       /* get the next attribute */
    switch (att_type(att->att)){/* and evaluate its type */
      case AT_FLT: att_inst(att->att)->f = 0;     break;
      case AT_INT: att_inst(att->att)->i = 0;     break;
      default    : att_inst(att->att)->n =
                     pt_rand(att->ptree, randfn); break;
    }                           /* clear numeric attributes and */
  }                             /* randomly instantiate nominal ones */
  return 0;                     /* return 'ok' */
}  /* gm_rand() */

/*--------------------------------------------------------------------*/

double gm_exec (GRAMOD *gm)
{                               /* --- determine prob./deg. of poss. */
  ATTID  i;                     /* loop variable */
  GMATT  *att;                  /* to traverse the attributes */
  double p = 1, t;              /* probability/possibility, buffer */

  assert(gm);                   /* check the function argument */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
    if (!(--att)->ptree) continue;
    t = pt_exec(att->ptree);    /* traverse the nominal attributes */
    if (t <= 0) return 0;       /* and compute the prob./poss. degree */
    if (gm->type == GM_PROB) p *= t;
    else if (t < p)          p  = t;
  }                             /* aggregate prob./poss. */
  return p;                     /* and return the result */
}  /* gm_exec() */

/*--------------------------------------------------------------------*/

int gm_execx (GRAMOD *gm, double res[])
{                               /* --- determine prob./deg. of poss. */
  ATTID  i, k, n;               /* loop variables, counter */
  INST   *inst;                 /* instance of attribute */
  GMATT  *att;                  /* to traverse the attributes */
  ATTID  *o;                    /* to traverse attribute identifiers */
  ATTID  *unks, *u;             /* unknown attribute ids. vector */
  ATTID  cnt;                   /* number of unknown attributes */
  double p;                     /* probability/possibility */

  assert(gm);                   /* check the function arguments */
  if (gm->type == GM_PROB) {    /* if probabilistic graphical model */
    if ((gm->order[0] < 0)      /* if no topological order exists */
    &&  (order(gm) != 0))       /* and it cannot be created either, */
      return -1;                /* abort the function */
  }                             /* (prob. model needs top. order) */

  /* --- collect unknown attributes --- */
  unks = u = gm->buf;           /* get the auxiliary vector */
  att  = gm->atts;              /* and traverse the attributes */
  for (i = 0; i < gm->attcnt; i++, att++) {
    if (!att->ptree) {          /* if attribute is not nominal, */
      att->prob = 1; continue;} /* set default prob./poss. degree */
    att->prob = -1;             /* invalidate prob./poss. degree */
    inst = att_inst(att->att);  /* get the attribute instance and */
    if (isnone(inst->i)) {      /* if the attribute value is unknown, */
      *u++ = i;                 /* note the attribute identifier */
      inst->i = att_valcnt(att->att) -1;
    }                           /* set the last value of the domain */
  }                             /* (first instantiation to try) */
  cnt = (ATTID)(u -unks);       /* compute number of unknown atts. */

  /* --- aggregate probabilities/possibility degrees --- */
  res[2] = res[0] = 0; n = 0;   /* init. the aggregates to compute */
  res[1] = INFINITY;            /* and the instantiation counter */
  do {                          /* value combination loop */
    p = 1;                      /* traverse attributes in top. order */
    o = (gm->order[0] >= 0) ? gm->order : NULL;
    for (i = gm->attcnt; --i >= 0; ) {
      att = gm->atts +((o) ? *o++ : i); /* get the condition data */
      if (att->prob < 0)        /* if the probability is invalid, */
        att->prob = pt_exec(att->ptree);        /* (re)compute it */
      if (att->prob <= 0) {    p  = 0; break; }
      if (gm->type == GM_PROB) p *= att->prob;
      else if (att->prob < p)  p  = att->prob;
    }                           /* compute value of instantiation */
    if (p < res[1]) res[1] = p; /* update minimum */
    if (p > res[2]) res[2] = p; /* and maximum */
    res[0] += p; n++;           /* and sum and count tuples */
    for (u = unks +(i = cnt); --i >= 0; ) {
      att = gm->atts +*--u;     /* invalidate prob./poss. of attrib. */
      att->prob = -1;           /* and traverse dependent attributes */
      for (o = att->deps +(k = att->depcnt); --k >= 0; )
        gm->atts[*--o].prob = -1;   /* invalidate prob./poss. degree */
      inst = att_inst(att->att);    /* get the attribute instance */
      if (--inst->i >= 0) break;
      inst->i = att_valcnt(att->att) -1;
    }                           /* compute next value combination */
  } while (i >= 0);             /* while there is another combination */
  if (n      >  0) res[0] /= (double)n;/* determine average value and */
  if (res[2] <= 0) res[1]  = 0; /* correct minimum if maximum is zero */

  /* --- clean up --- */
  for (u = unks +(i = cnt); --i >= 0; )
    att_inst(gm->atts[*--u].att)->i = NV_NOM;
  return 0;                     /* reset attributes to unknown values */
}  /* gm_execx() */             /* and return 'ok' */

/*----------------------------------------------------------------------
  Description Functions
----------------------------------------------------------------------*/

int gm_desc (GRAMOD *gm, FILE *file, int mode, int maxlen)
{                               /* --- describe a graphical model */
  int   len, l;                 /* loop variables for comments */
  ATTID i, k;                   /* loop variables, flag */
  GMATT *att;                   /* to traverse the attributes */

  assert(gm && file);           /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print header (as a comment) --- */
  if (mode & GM_TITLE) {        /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  graphical model\n");
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */
  mode &= ~GM_TITLE;            /* remove the title flag */

  /* --- print distributions --- */
  if (gm->type == GM_PROB)      /* if this is a probabilistic model */
    fprintf(file, "probnet = {\n");
  else                          /* if this is a possibilistic model */
    fprintf(file, "possnet = {\n  total = %g;\n\n", gm->tplcnt);
  att = gm->atts;               /* traverse the attributes */
  for (k = 0, i = gm->attcnt; --i >= 0; att++) {
    if (!att->ptree) continue;  /* skip non-nominal attributes */
    if (k++ > 0)                /* if this is not the first tree */
      fputc('\n', file);        /* that gets printed, skip a line */
    if (pt_desc(att->ptree, file, mode | PT_INDENT, maxlen) != 0)
      return -1;                /* print the prob./poss. tree */
  }
  fputs("};\n", file);          /* terminate the graphical model */

  /* --- print additional information (as a comment) --- */
  if (mode & GM_INFO) {         /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of attributes: %"ATTID_FMT, gm->attcnt);
    fprintf(file, "\n  number of conditions: %"ATTID_FMT, gm->concnt);
    if (gm->tplcnt > 0)
      fprintf(file, "\n  number of tuples    : %g", gm->tplcnt);
    fprintf(file, "\n  evaluation result   : %g\n", gm->eval);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print add. model information */
  }                             /* (as a comment) */
  return (ferror(file)) ? -1 : 0;
}  /* gm_desc() */              /* return the writing result */

/*----------------------------------------------------------------------
  Parse Functions
----------------------------------------------------------------------*/
#ifdef GM_PARSE

static int clists (GRAMOD *gm, SCANNER *scan)
{                               /* --- read candidate lists */
  int   t;                      /* buffer for a token */
  ATTID i, k, n;                /* attribute identifier, counter */
  ATTID cdd1, cdd2;             /* candidate identifiers */
  const char *s;                /* to access the token value */

  while ((scn_token(scan) == T_ID)  /* read the candidate lists */
  &&     (strcmp(scn_value(scan), "cand") == 0)) {
    SCN_NEXT(scan);             /* consume 'cand' */
    SCN_CHAR(scan, '(');        /* consume '(' */
    t = scn_token(scan);        /* check for an attribute name */
    if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
    i = as_attid(gm->attset, scn_value(scan));
    if (i < 0)                       SCN_ERROR(scan, E_UNKATT);
    if (gm->atts[i].pos)             SCN_ERROR(scan, E_DUPCDL);
    gm->atts[i].pos = -1;       /* mark the attribute as read */
    SCN_NEXT(scan);             /* and consume the attribute name */
    SCN_CHAR(scan, ')');        /* consume ')' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    SCN_CHAR(scan, '{');        /* consume '{' */
    k = 0;                      /* initialize the read counter */
    while (scn_token(scan) != '}') {     /* attribute read loop */
      if (++k > 1) {            /* if this is not the first att., */
        SCN_CHAR(scan, ','); }  /* consume ',' (separator) */
      t = scn_token(scan);      /* check for an attribute name */
      if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
      cdd1 = as_attid(gm->attset, scn_value(scan));
      if (cdd1 < 0)                    SCN_ERROR(scan, E_UNKATT);
      SCN_NEXT(scan);           /* consume the first candidate */
      t = scn_token(scan);      /* check for a range of atts. */
      s = scn_value(scan);      /* (that is: "att1 - att2") */
      if ((t != T_ID)           /* if no range is given */
      ||  (s[0] != '-') || (s[1] != '\0'))
        cdd2 = cdd1;            /* form a one element range */
      else {                    /* if a range is given */
        SCN_NEXT(scan);         /* consume '-' */
        t = scn_token(scan);    /* check for an attribute name */
        if ((t != T_ID) && (t != T_NUM)) SCN_ERROR(scan, E_ATTEXP);
        cdd2 = as_attid(gm->attset, s);
        if (cdd2 < 0)                    SCN_ERROR(scan, E_UNKATT);
        SCN_NEXT(scan);         /* consume the second candidate */
        if (cdd1 > cdd2)                 SCN_ERROR(scan, E_RANGE);
      }                         /* check the candidate range */
      for ( ; cdd1 <= cdd2; cdd1++) {
        if (cdd1 == i)          /* traverse condition candidates, */
          continue;             /* but ignore attribute itself */
        n = gm_cddadd(gm, i, cdd1);
        if (n >=  0) continue;  /* add a condition candidate */
        if (n >= -1) SCN_ERROR(scan, E_NOMEM);
        SCN_ERROR(scan, E_CAND, att_name(as_att(gm->attset, cdd1)));
      }                         /* if an error occurred, abort */
    }
    SCN_NEXT(scan);             /* consume '}' */
    SCN_CHAR(scan, ';');        /* consume ';' */
  }  /* while ((scn_token(scan) == T_ID) .. */
  return 0;                     /* return 'ok' */
}  /* clists() */

/*--------------------------------------------------------------------*/

static int cldflt (GRAMOD *gm)
{                               /* --- set default candidate lists */
  ATTID i, k;                   /* loop variables */
  GMATT *att;                   /* to traverse the attributes */
  int   type;                   /* attribute type */

  assert(gm);                   /* check the function argument */
  for (att = gm->atts +(i = gm->attcnt); --i >= 0; ) {
    type = att_type((--att)->att); /* traverse all attributes */
    if ((type     != AT_NOM)    /* if an attribute is not nominal */
    ||  (att->pos != 0)) {      /* or it is marked as read, */
      att->pos = 0; continue; } /* ignore the attribute */
    for (k = 0; k < gm->attcnt; k++)
      if ((k != i) && (gm_cddadd(gm, i, k) < 0))
        return -1;              /* add default candidate list */
  }                             /* (all other attributes) */
  return 0;                     /* return 'ok' */
}  /* cldflt() */

/*--------------------------------------------------------------------*/

static int setdeps (GRAMOD *gm)
{                               /* --- set list of dep. attributes */
  ATTID i, k;                   /* loop variables */
  PTREE *pt;                    /* prob./poss. tree of current att. */
  GMATT *con;                   /* to access the condition attribute */
  ATTID size;                   /* new dependent atts. vector size */
  ATTID *vec;                   /* new dependent atts. vector */

  assert(gm);                   /* check the function argument */
  for (i = gm->attcnt; --i >= 0; ) {
    pt = gm->atts[i].ptree;     /* traverse the attributes */
    if (!pt) continue;          /* with a condition tree */
    for (k = pt_concnt(pt); --k >= 0; ) {
      con = gm->atts +pt_conid(pt, k);
      if (con->depcnt >= con->depsize) {
        size  = con->depsize;   /* if dependent att. vector is full */
        size += (size > BLKSIZE) ? size >> 1 : BLKSIZE;
        vec   = (ATTID*)realloc(con->deps, (size_t)size *sizeof(ATTID));
        if (!vec) return -1;    /* enlarge the attribute vector */
        con->deps = vec; con->depsize = size;
      }                         /* set the new vector and its size */
      con->deps[con->depcnt++] = i;
    }                           /* the current attribute depends */
  }                             /* on its conditioning attributes */
  return 0;                     /* return 'ok' */
}  /* setdeps() */

/*--------------------------------------------------------------------*/

static int parse (ATTSET *attset, SCANNER *scan, int type, GRAMOD **gm)
{                               /* --- parse a graphical model */
  int   t = -1;                 /* type buffer */
  ATTID i;                      /* loop variable */
  GMATT *att;                   /* to traverse the attributes */
  PTREE *pt;                    /* buffer for prob./poss. tree */
  int   err = 0;                /* parse error status */

  assert(attset && scan && gm); /* check the function arguments */

  /* --- read header --- */
  if (scn_token(scan) == T_ID){ /* check for 'probnet' or 'possnet' */
    if      (strcmp(scn_value(scan), "probnet") == 0) t = GM_PROB;
    else if (strcmp(scn_value(scan), "possnet") == 0) t = GM_POSS;
  }                             /* determine the model type */
  if ((t < 0) && (type & GM_DFLT)) {
    *gm = gm_create(attset, type & GM_POSS);
    if (!*gm || (cldflt(*gm) != 0))
      SCN_ERROR(scan, E_NOMEM); /* if no proper token was found, */
    return 0;                   /* but the default flag is set, */
  }                             /* create a default model */
  if ((t < 0)                   /* if no proper token was found */
  ||  (!(type & GM_AUTO)        /* or not in automatic mode */
  &&   ((type ^ t) & GM_POSS))) /* and wrong token found, abort */
    SCN_ERROR(scan, E_STREXP, (type & GM_POSS) ? "possnet" : "probnet");
  if (type & GM_AUTO)           /* set the proper model type */
    type = (type & ~(GM_POSS|GM_AUTO)) | t;
  *gm = create(attset, type & GM_POSS);
  if (!*gm) SCN_ERROR(scan, E_NOMEM); /* create a graphical model */
  SCN_NEXT(scan);               /* consume 'probnet' or 'possnet' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */

  /* --- read total number of cases --- */
  if ((type & GM_POSS)          /* if this is a possibilistic model */
  &&  (scn_token(scan) == T_ID) /* check for 'total' or 'tplcnt' */
  &&  ((strcmp(scn_value(scan), "total")  == 0)
  ||   (strcmp(scn_value(scan), "tplcnt") == 0))) {
    SCN_NEXT(scan);             /* consume 'total' or 'tplcnt' */
    SCN_CHAR(scan, '=');        /* consume '=' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    (*gm)->tplcnt = atof(scn_value(scan));
    if ((*gm)->tplcnt < 0)      SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* consume the number of cases */
    SCN_CHAR(scan, ';');        /* consume ';' */
  }

  /* --- read candidate lists --- */
  if (type & GM_CDDS) {         /* if the candidates flag is set */
    while (1) {                 /* read loop (with recovery) */
      t = clists(*gm, scan);    /* read candidate lists */
      if (t == 0) break;        /* if no error occurred, abort */
      err = -1;                 /* otherwise set the error flag */
      if (t == E_NOMEM) break;  /* always abort on 'out of memory' */
      scn_recover(scan, ';', 0, 0, 0);
    }                           /* otherwise recover from the error */
    if (cldflt(*gm) != 0)       /* set default candidates */
      SCN_ERROR(scan, E_NOMEM); /* for all other attributes */
  }                             /* (make everything else a cdd.) */

  /* --- read prob./poss. distributions --- */
  while (scn_token(scan) != '}'){/* as long as there is another tree */
    pt = pt_parse(attset, scan, type);
    if (!pt) { err = -1; scn_recover(scan, ';', 0, 0, 0); continue; }
    att = (*gm)->atts +pt_attid(pt);
    att->ptree = pt;            /* read a prob./poss. tree and */
    att_setmark(att->att, 1);   /* mark the corresponding attribute */
    if ((type & GM_POSS)        /* for a possibilistic model */
    &&  ((*gm)->tplcnt >= 0))   /* set number of tuples for norm. */
      pt_setnorm(pt, (float)(*gm)->tplcnt);
    (*gm)->concnt += pt_concnt(pt);
  }                             /* sum the conditions */
  if (err) return err;          /* check for a parse error */
  (*gm)->type = type & GM_POSS; /* set the model type (prob./poss.) */
  for (att = (*gm)->atts, i = (*gm)->attcnt; --i >= 0; att++) {
    if (att->ptree              /* traverse the attributes */
    ||  (att_type(att->att) != AT_NOM))
      continue;                 /* skip read and numeric attributes */
    if (type & GM_ALL)          /* if all trees must be present */
      SCN_ERROR(scan, E_MISATT, att_name(att->att));
    att->ptree = pt_create(att->att, type);
    if (!att->ptree) SCN_ERROR(scan, E_NOMEM);
  }                             /* create missing prob./poss. trees */
  if ((i >= 0) || (setdeps(*gm) != 0))
    return -1;                  /* set list of dependent attributes */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return err;                   /* return the parse error status */
}  /* parse() */

/*--------------------------------------------------------------------*/

GRAMOD* gm_parse (ATTSET *attset, SCANNER *scan, int type)
{                               /* --- parse a graphical model */
  GRAMOD *gm = NULL;            /* created graphical model */

  assert(attset && scan);       /* check the function arguments */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (parse(attset, scan, type, &gm) != 0) {
    if (gm) gm_delete(gm, 0);   /* parse graphical model */
    return NULL;                /* if an error occurred, delete */
  }                             /* the graphical model and abort */
  return gm;                    /* return created graphical model */
}  /* gm_parse() */

#endif
