/*----------------------------------------------------------------------
  File    : ines.c
  Contents: Induction of NEtwork Structures (learning graphical models)
  Author  : Christian Borgelt
  History : 1995.11.03 file created
            1995.11.08 computation of probability distribution added
            1995.11.26 adapted to modified module attset
            1995.12.08 log file added, in and out file made global
            1996.01.18 adapted to modified function as_read
            1996.02.22 error message 'missing option argument' added
            1996.03.05 lower bound for improvement added
            1996.03.06 adapted to modified set of evaluation measures
            1996.03.17 adapted to modified parse function
            1996.11.22 options -b, -f, and -r (table structure) added
            1997.02.28 option -m (possibilistic model) added
            1997.03.08 list of evaluation measures added
            1998.02.02 option -u (unknown value characters) added
            1998.02.16 options -g, -G (merge, greedy merge) added
            1998.06.10 option -t (write a full tree) added
            1999.02.24 options -z (measure sensitivity)
                       and -p (prior/equiv. sample size) added
            1999.04.24 simplified using the new module 'io'
            2001.11.25 adapted to modified scanner functions
            2002.03.03 network induction functions moved to gramod2.c
            2002.04.11 simulated annealing added from file sian.c
            2003.01.16 optional approximation for poss. maximum removed
            2003.01.17 normalization of possibilistic measures added
            2003.08.16 slight changes in error message output
            2006.03.02 cond. independence tests construction added
            2006.03.04 adapted to modified induction functions
            2006.03.10 log file added to induction functions
            2006.12.20 adapted to simplified cond. indep. induction
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef GM_PARSE
#define GM_PARSE
#endif
#ifndef GM_EVAL
#define GM_EVAL
#endif
#ifndef GM_INDUCE
#define GM_INDUCE
#endif
#include "gramod.h"
#include "random.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "ines"
#define DESCRIPTION "induction of network structures"
#define VERSION     "version 4.1 (2014.10.24)         " \
                    "(c) 1995-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error */
#define E_SEARCH   (-10)        /* unknown search method */
#define E_MEASURE  (-11)        /* unknown measure */
#define E_COMBI    (-12)        /* invalid method/measure combination */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- measure information --- */
  int  code;                    /* measure code */
  char *name;                   /* name of the measure */
} MINFO;                        /* (measure information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const MINFO m_prob[] = { /* --- probabilistic measures */
  { PT_NONE,     "none"     },  /* no measure */
  { PT_INFGAIN,  "infgain"  },  /* information gain */
  { PT_INFGBAL,  "infgbal"  },  /* balanced information gain */
  { PT_INFGR,    "infgr"    },  /* information gain ratio */
  { PT_INFSGR1,  "infsgr1"  },  /* sym. information gain ratio 1 */
  { PT_INFSGR2,  "infsgr2"  },  /* sym. information gain ratio 2 */
  { PT_QIGAIN,   "qigain"   },  /* quadratic information gain */
  { PT_QIGBAL,   "qigbal"   },  /* balanced quad. information gain */
  { PT_QIGR,     "qigr"     },  /* quadratic information gain ratio */
  { PT_QISGR1,   "qisgr1"   },  /* sym. quad. info. gain ratio 1 */
  { PT_QISGR2,   "qisgr2"   },  /* sym. quad. info. gain ratio 2 */
  { PT_GINI,     "gini"     },  /* Gini index */
  { PT_GINISYM,  "ginisym"  },  /* symmetric Gini index */
  { PT_GINIMOD,  "ginimod"  },  /* modified Gini index */
  { PT_RELIEF,   "relief"   },  /* relief measure */
  { PT_WDIFF,    "wdiff"    },  /* sum of weighted differences */
  { PT_CHI2,     "chi2"     },  /* chi^2 measure */
  { PT_CHI2NRM,  "chi2nrm"  },  /* normalized chi^2 measure */
  { PT_WEVID,    "wevid"    },  /* weight of evidence */
  { PT_RELEV,    "relev"    },  /* relevance */
  { PT_BDM,      "bdm"      },  /* Bayesian-Dirichlet / K2 metric */
  { PT_BDMOD,    "bdmod"    },  /* modified BD / K2 metric */
  { PT_RDLREL,   "rdlrel"   },  /* red. of desc. length (rel. freq.) */
  { PT_RDLABS,   "rdlabs"   },  /* red. of desc. length (abs. freq.) */
  { PT_STOCO,    "stoco"    },  /* stochastic complexity */
  { -1,          NULL       }   /* sentinel */
};

static const MINFO m_poss[] = { /* --- possibilistic measures */
  { PT_SPCGAIN,  "spcgain"  },  /* specificity gain */
  { PT_SPCGBAL,  "spcgbal"  },  /* balanced specificity gain */
  { PT_SPCGR,    "spcgr"    },  /* specificity gain ratio */
  { PT_SPCSGR1,  "spcsgr1"  },  /* sym. specificity gain ratio 1 */
  { PT_SPCSGR2,  "spcsgr2"  },  /* sym. specificity gain ratio 2 */
  { PT_WDIFF,    "wdiff"    },  /* weighted (absolute) difference */
  { PT_CHI2,     "chi2"     },  /* possibilistic chi^2 measure */
  { PT_MUTSPC,   "mutspc"   },  /* mutual specificity */
  { -1,          NULL       }   /* sentinel */
};

static const MINFO sminfo[] = { /* search method information */
  { GM_NONE,     "none"     },  /* no search, only param. estim. */
  { GM_OWST,     "owst"     },  /* optimum weight spanning tree */
  { GM_EXTST,    "extst"    },  /* extension of spanning tree */
  { GM_TOPORD,   "topord"   },  /* greedy on topological order */
  { GM_NOLOOP,   "noloop"   },  /* greedy avoiding loops */
  { GM_UNRES,    "unres"    },  /* unrestricted greedy */
  { GM_CITEST,   "citest"   },  /* cond. independence tests, directed */
  { GM_CIHEUR,   "ciheur"   },  /* ditto, without strict thinning */
  { GM_CIUNDIR,  "ciundir"  },  /* cond. independence tests, undir. */
  { GM_CIUDNIT,  "ciudnit"  },  /* ditto, without intermed. thinning */
  { GM_SIAN,     "sian"     },  /* hypertree simulated annealing */
  { -1,          NULL       }   /* sentinel */
};

static const char *errmsgs[] = {/* error messages */
  /* E_NONE      0 */ "no error",
  /* E_NOMEM    -1 */ "not enough memory",
  /* E_FOPEN    -2 */ "cannot open file %s",
  /* E_FREAD    -3 */ "read error on file %s",
  /* E_FWRITE   -4 */ "write error on file %s",
  /* E_STDIN    -5 */ "double assignment of standard input",
  /* E_OPTION   -6 */ "unknown option -%c",
  /* E_OPTARG   -7 */ "missing option argument",
  /* E_ARGCNT   -8 */ "wrong number of arguments",
  /* E_PARSE    -9 */ "parse error(s) on file %s",
  /* E_SEARCH  -10 */ "unknown search method %s",
  /* E_MEASURE -11 */ "unknown evaluation measure %s",
  /* E_COMBI   -12 */ "invalid method/measure combination",
  /*           -13 */ "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain file) */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static TABLE   *table  = NULL;  /* table to induce network from */
static GRAMOD  *gramod = NULL;  /* induced graphical model */
static FILE    *out    = NULL;  /* output file */
static FILE    *logout = NULL;  /* log file (optional) */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (table)  tab_delete(table, 0); \
  if (gramod) gm_delete(gramod, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out    && (out    != stdout)) fclose(out); \
  if (logout && (logout != stderr)) fclose(logout);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void help (void)
{                               /* --- print help on methods/measures */
  const MINFO *m;               /* to travserse the measure info */

  fprintf(stderr, "\n");        /* terminate the startup message */
  printf("list of search methods (option -s#)\n");
  for (m = sminfo; m->name; m++)
    printf("  %-9s  %s\n",   m->name, gm_sname(m->code));
  printf("\nlist of evaluation measures (option -e#)\n");
  printf("measures for probabilistic networks\n");
  printf("  name       measure\n");
  for (m = m_prob; m->name; m++)
    printf("  %-9s  %s%s\n", m->name, pt_mname(PT_PROB, m->code),
           (pt_minfo(PT_PROB, m->code) == PT_SYM) ? " *" : "");
  printf("\nmeasures for possibilistic networks\n");
  printf("  name       measure\n");
  for (m = m_poss; m->name; m++)
    printf("  %-9s  %s%s\n", m->name, pt_mname(PT_POSS, m->code),
           (pt_minfo(PT_POSS, m->code) == PT_SYM) ? " *" : "");
  printf("\nThe search methods owst, extst, and citest"
         " can only be used with\n"
         "symmetric evaluation measures, ");
  printf("which are marked with a *.\n");
  exit(0);                      /* abort porgram */
}  /* help() */

/*--------------------------------------------------------------------*/

static int code (const MINFO *tab, const char *name)
{                               /* --- get measure code */
  assert(tab && name);          /* check the function arguments */
  for ( ; tab->name; tab++)     /* look up name in table */
    if (strcmp(tab->name, name) == 0)
      return tab->code;         /* return the measure code */
  return -1;                    /* or an error indicator */
}  /* code() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domains file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_net  = NULL;      /* name of network file */
  CCHAR   *fn_log  = NULL;      /* name of log file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *sname   = NULL;      /* name of search method */
  CCHAR   *mname   = NULL;      /* name of evaluation measure */
  int     type     = GM_PROB;   /* type of graphical model */
  int     search   = 0;         /* search method */
  int     measure  = 0;         /* evaluation measure */
  double  params[] = { 0, 1 };  /* evaluation measure parameters */
  double  minimp   = 0;         /* minimal evaluation improvement */
  double  thresh   = 0.1;       /* threshold for cond. independence */
  double  mrgimp   = 0;         /* minimal imp. for distrib. merging */
  int     local    = -1;        /* flag for local structure learning */
  ATTID   maxcon   = 2;         /* maximum number of conditions */
  ATTID   maxtest  = 64;        /* maximum number of parallel tests */
  int     trials   = 1000;      /* number of trials */
  double  keep     = 0.8;       /* fraction of hyperedges to keep */
  double  qrfrac   = 0.001;     /* final fraction of quality range */
  double  ppwgt    = 0.0;       /* parameter penalty weight */
  int     mode     = AS_ATT|AS_NOEXT;  /* table read mode */
  int     desc     = GM_TITLE|GM_INFO; /* description mode */
  int     maxlen   = 0;         /* maximal output line length */
  long    seed;                 /* seed for random numbers */
  double  eval;                 /* evaluation result */
  ATTID   m;                    /* number of attributes */
  TPLID   n;                    /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */
  seed    = (long)time(NULL);   /* and get a default seed value */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] domfile "
           "[-d|-h hdrfile] tabfile netfile [logfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-m       possibilistic mode "
                    "(default: probabilistic mode)\n");
    printf("-s#      search method      "
                    "(default: none, only parameter estimation)\n");
    printf("-e#      evaluation measure (default: infgain/spcgain)\n");
    printf("-!       print a list of search methods "
                    "and evaluation measures\n");
    printf("-z#      sensitivity parameter                  "
                    "(default: %g)\n", params[0]);
    printf("         (only used for measures wdiff, bdm, bdmod, "
                    "rdlrel, and rdlabs)\n");
    printf("-p#      prior (if positive) or "
                    "equivalent sample size (if negative)\n");
    printf("         (only used for measures bdm and bdmod; "
                    "default: %g)\n", params[1]);
    printf("-c#      maximum number of conditions "
                    "to select per attribute\n");
    printf("         or maximum size of a hyperedge minus 1 "
                    "(default: %"ATTID_FMT")\n", maxcon);
    printf("-i#      minimum improvement/indep. threshold   "
                    "(default: %g/%g)\n", minimp, thresh);
    printf("-g#/G#   greedy merging of conditional dists.   "
                    "(default: no merging)\n");
    printf("-y#      maximum number of parallel tests       "
                    "(default: %"ATTID_FMT")\n", maxtest);
    printf("-t#      number of trials (simulated annealing) "
                    "(default: %d)\n", trials);
    printf("-k#      percentage of hyperedges to keep "
                    "(default: %g%%)\n", keep *100);
    printf("-j#      final fraction of quality range        "
                    "(default: %g)\n", qrfrac);
    printf("-w#      penalty weight for simulated annealing "
                    "(default: %g)\n", ppwgt);
    printf("-S#      seed for random number generator       "
                    "(default: time)\n");
    printf("-x       print a full probability/possibility tree "
                    "(all branches)\n");
    printf("-q       print relative frequencies (in percent)\n");
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("-r#      record     separators                  "
                    "(default: \"\\n\")\n");
    printf("-f#      field      separators                  "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank      characters                  "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment    characters                  "
                    "(default: \"#\")\n");
    printf("domfile  file containing domain descriptions "
                    "(in topological order)\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("netfile  file to write induced network structure to\n");
    /* unused lowercase letters: o, v */
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case '!': help();                               break;
          case 'm': type      = GM_POSS;                  break;
          case 's': optarg    = &sname;                   break;
          case 'e': optarg    = &mname;                   break;
          case 'z': params[0] =        strtod(s, &s);     break;
          case 'p': params[1] =        strtod(s, &s);     break;
          case 'c': maxcon    = (ATTID)strtol(s, &s, 0);  break;
          case 'i': minimp    = thresh = strtod(s, &s);   break;
          case 'g': mrgimp    = strtod(s, &s); local = 0; break;
          case 'G': mrgimp    = strtod(s, &s); local = 1; break;
          case 'y': maxtest   = (ATTID)strtol(s, &s, 0);  break;
          case 't': trials    = (int)  strtol(s, &s, 0);  break;
          case 'k': keep      = 0.01  *strtod(s, &s);     break;
          case 'j': qrfrac    =        strtod(s, &s);     break;
          case 'w': ppwgt     =        strtod(s, &s);     break;
          case 'S': seed      =        strtol(s, &s, 0);  break;
          case 'x': desc     |= GM_FULL;                  break;
          case 'q': desc     |= GM_REL;                   break;
          case 'l': maxlen    = (int)  strtol(s, &s, 0);  break;
          case 'r': optarg    = &recseps;                 break;
          case 'f': optarg    = &fldseps;                 break;
          case 'b': optarg    = &blanks;                  break;
          case 'C': optarg    = &comment;                 break;
          case 'u': optarg    = &nullchs;                 break;
          case 'n': mode     |= AS_WEIGHT;                break;
          case 'd': mode     |= AS_DFLT;                  break;
          case 'h': optarg    = &fn_hdr;                  break;
          default : error(E_OPTION, *(--s));              break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_net = s;      break;
        case  3: fn_log = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the program arguments */
  if ((k != 3) && (k != 4)) error(E_ARGCNT);
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_dom || !*fn_dom) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  search  = (sname) ? code(sminfo, sname) : GM_NONE;
  if (search < 0) error(E_SEARCH);
  measure = (type == PT_PROB)   /* get search method and measure */
          ? ((mname) ? code(m_prob, mname) : PT_INFGAIN)
          : ((mname) ? code(m_poss, mname) : PT_SPCGAIN);
  if (measure < 0) error(E_MEASURE, mname);
  if (((search == GM_OWST)      /* for the tree search methods */
  ||   (search == GM_EXTST)     /* and the methods based on */
  ||  ((search >= GM_CITEST)    /* conditional independence tests */
  &&   (search <= GM_CIUDNIT))) /* the measure must be symmetric */
  &&  !(pt_minfo(type, measure) & PT_SYM))
    error(E_COMBI);             /* check method/measure combination */
  if (trials <= 0) trials = 1;  /* adapt number of trials */
  if (maxcon <  0) maxcon = 0;  /* and maximum number of conditions */
  if (keep   <  0) keep   = 0;  /* check and adapt the percentage */
  if (keep   >  1) keep   = 1;  /* of hyperedges to keep */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan));

  /* --- parse existing graphical model --- */
  gramod = gm_parse(attset, scan, type|GM_CDDS|GM_DFLT);
  if (!gramod || !scn_eof(scan, 1)) /* parse or create */
    error(E_PARSE, scn_name(scan)); /* a graphical model */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  table = tab_create("table", attset, tpl_delete);
  if (!table) error(E_NOMEM);   /* create a data table */
  k = tab_read(table, tread, mode);
  if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(table);        /* get the number of attributes */
  n = tab_tplcnt(table);        /* and the number of data tuples */
  w = tab_tplwgt(table);        /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- reduce table --- */
  fprintf(stderr, "reducing table ... ");
  tab_reduce(table);            /* reduce the table for speed up */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- induce graphical model --- */
  fprintf(stderr, "inducing network ... ");
  if (fn_log) {                 /* if a log file name is given */
    logout = (!*fn_log) ? stderr : fopen(fn_log, "w");
    if (!logout) error(E_FOPEN, fn_log);
  }                             /* open the log file for writing */
  if      ((search == GM_OWST)  /* if optimum weight tree induction */
  ||       (search == GM_EXTST)) {
    eval = gm_tree  (gramod, table, search, measure, params,
                     thresh, maxtest, logout); }
  else if ((search == GM_TOPORD)/* if greedy condition selection */
  ||       (search == GM_NOLOOP)/* (with different restrictions) */
  ||       (search == GM_UNRES)) {
    eval = gm_consel(gramod, table, search, measure, params,
                     minimp, local, mrgimp, maxcon, maxtest, logout); }
  else if ((search >= GM_CITEST)/* if cond. indep. tests constr. */
  &&       (search <= GM_CIUDNIT)) {
    eval = gm_citest(gramod, table, search, measure, params,
                     thresh, local, mrgimp, maxcon, maxtest, logout); }
  else if  (search == GM_SIAN){ /* if to do simulated annealing */
    rseed((unsigned int)seed);  /* init. the random number generator */
    eval = gm_sian  (gramod, table, trials, maxcon+1,
                     keep, qrfrac, ppwgt, 0, drand); }
  else {                        /* no search, parameter induction */
    eval = gm_params(gramod, table, measure, params);
    if (!isnan(eval) && (local >= 0))
      eval = gm_local(gramod, local, mrgimp);
  }                             /* (induce a graphical model) */
  if (isnan(eval)) error(E_NOMEM);
  if (logout && (logout != stderr)) {/* if log not written to stderr, */
    i = fclose(logout); logout = NULL;          /* close the log file */
    if (i) error(E_FWRITE, fn_log);
  }                             /* print network evaluation */
  fprintf(stderr, "[evaluation: %g] done.\n", eval);

  /* --- write network structure --- */
  t = clock();                  /* start timer, open output file */
  if (fn_net && (strcmp(fn_net, "-") == 0)) fn_net = "";
  if (fn_net && *fn_net) { out = fopen(fn_net, "w"); }
  else                   { out = stdout; fn_net = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_net);
  if (!out) error(E_FOPEN, fn_net);
  if (as_desc(attset, out, AS_TITLE, maxlen) != 0)
    error(E_FWRITE, fn_net);    /* write the domain definitions */
  fputc('\n', out);             /* leave one line empty */
  if (gm_desc(gramod, out, desc, maxlen) != 0)
    error(E_FWRITE, fn_net);    /* write the network structure */
  if (((out == stdout) ? fflush(out) : fclose(out)) != 0)
    error(E_FWRITE, fn_net);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)",  gm_attcnt(gramod));
  fprintf(stderr, "/%"ATTID_FMT" condition(s)]", gm_concnt(gramod,-1));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
