/*----------------------------------------------------------------------
  File    : ptree5.c
  Contents: probability/possibility tree management
            local structure learning functions
  Author  : Christian Borgelt
  History : 1995.10.22 file created as ctree.c
            2002.01.12 local structure learning moved to separate file
            2002.07.04 function logGamma() made external
            2002.07.17 function tables redesigned, nsp() moved to ptree3
            2002.07.18 optimizations for independent gains added
            2003.01.17 normalization of possibilistic measures added
            2003.01.18 bug in computation of balanced spc. gain fixed
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "gamma.h"
#include "ptree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define M_PI   3.14159265358979323846  /* \pi   */
#define M_LN2  0.69314718055994530942  /* ln(2) */

/* --- flags for leaf merger evaluation --- */
#define MRG_INDEP     0         /* gain is independent of mergers */
#define MRG_LOCAL     1         /* local  recomputation necessary */
#define MRG_GLOBAL    2         /* global recomputation necessary */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double EVALFN (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2);

typedef struct {                /* --- function table element */
  EVALFN *evalfn;               /* leaf merger evaluation function */
  int    type;                  /* evaluation type, e.g. MRG_INDEP */
} FTE;                          /* (function table element) */

typedef struct {                /* --- merge table --- */
  EVALFN *evalfn;               /* leaf merger evaluation function */
  double minimp;                /* minimal improvement */
  VALID  leafcnt;               /* number of leaves in the tree */
  VALID  rowcnt;                /* current number of table rows */
  VALID  *bids;                 /* indices of best mergers per row */
  double **evals;               /* evaluations of mergers */
  PTLEAF *leaves[1];            /* list of leaf nodes */
} MRGTAB;                       /* (merge table) */

/*----------------------------------------------------------------------
  Evaluation Functions
----------------------------------------------------------------------*/

static double info (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- Shannon information change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n1, n2, n;             /* leaf totals */
  double s, t;                  /* sum of entropy terms */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* simply copy N H_i (no change) */
  n1 = leaf1->total;            /* get the two leaf totals, sum them, */
  n2 = leaf2->total; n = n1+n2; /* and update N H_j (marginal dist.) */
  t  = n1*log(n1) +n2*log(n2) -n*log(n);
  pt->res1[1] = pt->res0[1] +((t < 0) ? t : 0);
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the counters */
  for (s = 0; --i >= 0; ) {     /* of the two leaves and */
    c = *--c1 + *--c2;          /* compute the entropy term */
    if (c > 0) s += c*log(c);   /* of a merged leaf, then */
  }                             /* update N H_ij (joint distribution) */
  t = leaf1->eval +leaf2->eval -s;
  pt->res1[2] = pt->res0[2] +((t < 0) ? t : 0);
  pt->res1[3] = pt->res0[3];    /* note the normalization factors */
  i = pt->leafcnt -1;           /* (for one leaf less) */
  pt->res1[4] = (i > 1) ? 1/(pt->total *log((double)i)) : 0;
  return s;                     /* return the leaf evaluation */
}  /* info() */

/*--------------------------------------------------------------------*/

static double quad (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- quadratic information change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n1, n2, n;             /* leaf totals */
  double s, t;                  /* sum of entropy terms */

  assert(pt && leaf1 && leaf2); /* check the function arguments */
  pt->res1[0] = pt->res0[0];    /* simply copy N^2/2 H^2_i */
  n1 = leaf1->total;            /* get the leaf totals, */
  n2 = leaf2->total; n = n1+n2; /* sum them (total of merged leaf), */
  t  = n1*n1 +n2*n2 -n*n;       /* and update N^2/2 H^2_j */
  pt->res1[1] = pt->res0[1] +((t < 0) ? t : 0);
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  for (s = 0; --i >= 0; ) {     /* of the two leaves and */
    c = *--c1 + *--c2;          /* compute the entropy term */
    s += c *c;                  /* of a merged leaf, */
  }                             /* then update N^2/2 H^2_ij */
  t = leaf1->eval +leaf2->eval -s;
  pt->res1[2] = pt->res0[2] +((t < 0) ? t : 0);
  pt->res1[3] = pt->res0[3];    /* note the normalization factors */
  i = pt->leafcnt -1;           /* (for one leaf less) */
  pt->res0[4] = (i > 0) ? 1/(s *(1 -1/(double)i)) : 0;
  return s;                     /* return the leaf evaluation */
}  /* quad() */

/*--------------------------------------------------------------------*/

static double spec (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- specificity change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *b, *c1, *c2;          /* to traverse the buffers/counters */

  assert(pt && (pt->total > 0));/* check the function argument */
  pt->res1[0] = pt->res0[0];    /* simply copy S_i = nsp(child) */
  lvl = pt->levels +pt->concnt; /* get the leaf level and */
  b = pt->buf;                  /* process distribution on conditions */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    if ((leaf != leaf1) && (leaf != leaf2) && (leaf->total > 0))
      *b++ = (WEIGHT)leaf->total;  /* collect the leaf totals */
  *b++ = (WEIGHT)(leaf1->total +leaf2->total);
  pt->res1[1] = pt_nsp(pt->buf, (VALID)(b -pt->buf));
  b = pt->buf;                  /* process the joint distribution */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if ((leaf == leaf1) || (leaf == leaf2) || (leaf->total <= 0))
      continue;                 /* traverse nonempty leaves */
    for (c1 = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      if (*--c1 > 0) *b++ = *c1;/* copy nonzero counters to buffer */
  }                             /* and compute S_ij from them */
  c1 = leaf1->cnts +(i = lvl->cnt);
  c2 = leaf2->cnts + i;         /* add the counters for the merger */
  while (--i >= 0) *b++ = *--c1 + *--c2;
  pt->res1[2] = pt_nsp(pt->buf, (VALID)(b -pt->buf));
  pt->res1[3] = pt->res0[3];    /* note normalization factors */
  i = pt->leafcnt -1;           /* (for one leaf less) */
  pt->res1[4] = (i > 1) ? M_LN2/log((double)i) : 1;
  return 0;                     /* return a dummy evaluation */
}  /* spec() */

/*--------------------------------------------------------------------*/

static double gini1 (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- Gini index 1 change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n1,  n2,  n;           /* leaf counter totals */
  double s, t;                  /* sum of squared frequencies */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* simply copy sum_i N_i^2 */
  n1 = leaf1->total;            /* get the leaf totals, */
  n2 = leaf2->total; n = n1+n2; /* sum them (total of merged leaf), */
  t  = n*n -n1*n1 -n2*n2;       /* and update sum_j N_j^2 */
  pt->res1[1] = pt->res0[1] +((t > 0) ? t : 0);
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  for (s = 0; --i >= 0; ) {     /* of the two leaves and */
    c = *--c1 + *--c2;          /* compute the squared frequencies */
    s += c *c;                  /* of a merged leaf, */
  }                             /* then update sum_ij N_ij^2 */
  t = s -leaf1->eval -leaf2->eval;
  pt->res1[2] = pt->res0[2] +((t > 0) ? t : 0);
  pt->res1[3] = pt->res0[3] +(s/n -leaf1->eval/n1 -leaf2->eval/n2);
  return s;                     /* return the leaf evaluation */
}  /* gini1() */

/*--------------------------------------------------------------------*/

static double gini2 (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- Gini index 2 change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n1,  n2,  n;           /* leaf counter totals */
  double s, t;                  /* sum of squared frequencies */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* simply copy sum_i N_i^2 */
  n1 = leaf1->total;            /* get the leaf totals, */
  n2 = leaf2->total; n = n1+n2; /* sum them (total of merged leaf), */
  t  = n*n -n1*n1 -n2*n2;       /* and update sum_j N_j^2 */
  pt->res1[1] = pt->res0[1] +((t > 0) ? t : 0);
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters and */
  b   = lvl->buf +i +i;         /* the second evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the squared frequencies */
    c = *--c1 + *--c2;          /* of a merged leaf and update */
    s    += c *= c;             /* the buffered squared frequencies */
    *--b += c -*c1**c1 -*c2**c2;
  }                             /* update the sums s_ij and w_ij */
  t = s -leaf1->eval -leaf2->eval;
  pt->res1[2] = pt->res0[2] +((t > 0) ? t : 0);
  pt->res1[3] = pt->res0[3] +(s/n -leaf1->eval/n1 -leaf2->eval/n2);
  for (n = 0, c1 = b +(i = lvl->cnt); --i >= 0; )
    n += *--c1 / *--b;          /* completely recompute */
  pt->res1[4] = n;              /* w_ji = sum_i (1/N_i) sum_j N_ij^2 */
  return s;                     /* return the leaf evaluation */
}  /* gini2() */

/*--------------------------------------------------------------------*/

static double bdm (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- Bayesian-Dirichlet change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2;              /* to traverse the counters */
  double s, t;                  /* sum of BD metric terms, buffer */
  double p, a;                  /* prior and sensitivity */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* copy the reference evaluation */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  a   = (pt->params[0] > -1) ? pt->params[0]+1 : 1;
  a  *= a;                      /* get the sensitivity parameter */
  p   = (pt->params[1] != 0) ? pt->params[1]   : 1;
  if (p < 0)                    /* get the prior/equiv. sample size */
    p = (-p /(double)lvl->cnt /(double)pt->fullcnt)
      * ((double)leaf1->pathcnt +(double)leaf2->pathcnt);
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters and */
  for (s = 0; --i >= 0; )       /* sum the BD terms for the leaf */
    s += logGamma((*--c1 +*--c2) *a +p);
  t  = (double)lvl->cnt *p;     /* compute the leaf term and update */
  s += logGamma(t) -(double)lvl->cnt *logGamma(p)  /* the BD metric */
     - logGamma((leaf1->total +leaf2->total) *a +t);
  pt->res1[1] = pt->res0[1] +(s -leaf1->eval -leaf2->eval);
  return s;                     /* return the leaf evaluation */
}  /* bdm() */

/*--------------------------------------------------------------------*/

static double bdmod (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- mod. Bayesian-Dirichlet change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n, t, s;               /* sum of BD metric terms, buffer */
  double p, a;                  /* prior and sensitivity */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* copy the reference evaluation */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  a   = (pt->params[0] > -1) ? pt->params[0]+1 : 1;
  if (a < 0) a = 0;             /* get the sensitivity parameter */
  p   = (pt->params[1] != 0) ? pt->params[1]   : 1;
  if (p < 0)                    /* get the prior/equiv. sample size */
    p = (-p /(double)lvl->cnt /(double)pt->fullcnt)
      * ((double)leaf1->pathcnt +(double)leaf2->pathcnt);
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters and */
  for (s = 0; --i >= 0; ) {     /* sum the BD terms for the counters */
    c = *--c1 +*--c2; t = c *a +p; s += logGamma(t +c) -logGamma(t); }
  n  = leaf1->total +leaf2->total;
  t  = n *a +(double)lvl->cnt *p; /* compute the leaf term and update */
  s += logGamma(t) -logGamma(t +n);         /* the modified BD metric */
  pt->res1[1] = pt->res0[1] +(s -leaf1->eval -leaf2->eval);
  return s;                     /* return the leaf evaluation */
}  /* bdmod() */

/*--------------------------------------------------------------------*/

static double dlrel (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- desc. length change (rel.frq.) */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n, s, lG;              /* sum of description length terms */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* copy the reference evaluation */
  pt->res1[1] = pt->res0[1];    /* (data and model desc. length) */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters and */
  for (s = 0; --i >= 0; ) {     /* sum the entropy terms for the leaf */
    c = *--c1 +*--c2; if (c > 0) s += c *log(c); }
  n  = leaf1->total +leaf2->total;  /* evaluate the merged leaf */
  lG = logGamma(n+1); s  = n *log(n) -s;
  pt->res1[2] = pt->res0[2] +(s -leaf1->eval -leaf2->eval);
  pt->res1[3] = pt->res0[3]     /* update the description lengths */
              + logGamma(n +(double)lvl->cnt) -lG
              + logGamma(   (double)lvl->cnt)
              - logGamma(leaf1->total +(double)lvl->cnt)
              + logGamma(leaf1->total +1)
              - logGamma(leaf2->total +(double)lvl->cnt)
              + logGamma(leaf2->total +1);
  return s;                     /* return the leaf evaluation */
}  /* dlrel() */

/*--------------------------------------------------------------------*/

static double dlabs (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- desc. length change (abs.frq.) */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c;           /* to traverse the counters */
  double n, s, lG;              /* sum of description length terms */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  pt->res1[0] = pt->res0[0];    /* copy the reference evaluation */
  pt->res1[1] = pt->res0[1];    /* (data and model desc. length) */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters and */
  for (s = 0; --i >= 0; ) {     /* compute bits for code book page */
    c = *--c1 +*--c2; s += logGamma(c+1); }
  n  = leaf1->total +leaf2->total;
  lG = logGamma(n+1); s = lG -s;/* evaluate the merged leaf */
  pt->res1[2] = pt->res0[2] +(s -leaf1->eval -leaf2->eval);
  pt->res1[3] = pt->res0[3]     /* update the description lengths */
              + logGamma((double)lvl->cnt +n) -lG
              + logGamma((double)lvl->cnt)
              - logGamma(leaf1->total +(double)lvl->cnt)
              + logGamma(leaf1->total +1)
              - logGamma(leaf2->total +(double)lvl->cnt)
              + logGamma(leaf2->total +1);
  return s;                     /* return the leaf evaluation */
}  /* dlabs() */

/*--------------------------------------------------------------------*/

static double stoco (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- stochastic complexity change */
  PTLVL  *lvl;                  /* leaf level description */
  double n1, n2, n;             /* leaf counter totals */
  double s;                     /* sum of squared frequencies */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  s   = info(pt,leaf1,leaf2);   /* compute the entropies */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n1  = leaf1->total;           /* get the leaf totals and */
  n2  = leaf2->total; n = n1+n2;/* update the penalty term */
  pt->res1[3] = pt->res0[3] +(log(0.5*n) -log(0.5*n1) -log(0.5*n2))
                            *0.5 *((double)lvl->cnt -1);
  pt->res1[4] = (2 -(double)pt->leafcnt)
              * log(pow(M_PI, 0.5 *(double)lvl->cnt))
                   - logGamma(0.5 *(double)lvl->cnt);
  return s;                     /* return the leaf evaluation */
}  /* stoco() */

/*--------------------------------------------------------------------*/

static double wdiff (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- weighted difference change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n, s, t;               /* sum of differences, buffer */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    c = *--c1 + *--c2;          /* of a merged leaf, i.e., */
    t = *--b *n - c *pt->total; /* for summed counters */
    if      (pt->params[0] == 1) s += c *fabs(t);
    else if (pt->params[0] == 2) s += c *t *t;
    else                         s += c *pow(fabs(t), pt->params[0]);
  }                             /* update the sum of differences */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* wdiff() */

/*--------------------------------------------------------------------*/

static double chi2 (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- chi^2 measure change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, *b;          /* to traverse the counters */
  double n, s, t, p;            /* sum of chi^2 terms, buffers */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    --c1; --c2; p = *--b *n;    /* of a merged leaf, i.e., */
    if (p <= 0) continue;       /* for summed counters */
    t  = p - (*c1 + *c2) *pt->total;
    s += t *t /p;               /* sum the chi^2 terms and */
  }                             /* update the chi^2 measure */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* chi2() */

/*--------------------------------------------------------------------*/

static double evid (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- weight of evidence change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n, s, x, y, z;         /* sum of evidence terms, buffers */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    c = *--c1 + *--c2;          /* of a merged leaf, i.e., */
    z = (double)c * *--b;       /* for summed counters */
    x = (double)c *pt->total -z; if (x == 0) continue;
    y = (double)*b *n        -z; if (y == 0) continue;
    s += *b *fabs(log(x/y));    /* sum the odds fractions weighted */
  }                             /* with the marginal probability */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* evid() */

/*--------------------------------------------------------------------*/

static double relev (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- relevance change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, *b;          /* to traverse the counters */
  double s, m, t;               /* sum of relevance terms, buffer */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = m = 0; --i >= 0; ) { /* compute the evaluation */
    --c1; --c2;                 /* of a merged leaf, i.e., */
    if (*--b <= 0) continue;    /* for summed counters */
    t = (double)(*c1 +*c2) / *b;
    if (t <= m) { s += t; } else { s += m; m = t; }
  }                             /* update the total relevance */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* relev() */

/*--------------------------------------------------------------------*/

static double pdiff (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- weighted difference change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n, s, t;               /* sum of differences, buffer */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    c = *--c1 + *--c2;          /* of a merged leaf */
    t = ((*--b < n) ? *b : n) - c;
    if      (pt->params[0] == 1) s += c *fabs(t);
    else if (pt->params[0] == 2) s += c *t *t;
    else                         s += c *pow(fabs(t), pt->params[0]);
  }                             /* update the sum of differences */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* pdiff() */

/*--------------------------------------------------------------------*/

static double pchi2 (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- possibilistic chi^2 change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n, s, t, p;            /* sum of chi^2 terms, buffers */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    c = *--c1 + *--c2;          /* of a merged leaf, i.e., */
    p = (*--b < n) ? *b : n;    /* for summed counters */
    if (p <= 0) continue;
    t = p -c; s += t *t /p;     /* sum the chi^2 terms and */
  }                             /* update the chi^2 measure */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2;  /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* pchi2() */

/*--------------------------------------------------------------------*/

static double mutspc (PTREE *pt, PTLEAF *leaf1, PTLEAF *leaf2)
{                               /* --- mutual specificity change */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  WEIGHT *c1, *c2, c, *b;       /* to traverse the counters */
  double n, s, t;               /* sum of difference terms, buffer */

  assert(pt && leaf1 && leaf2   /* check the function arguments */
  &&    (leaf1->total > 0) && (leaf2->total > 0));
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  n   = leaf1->total +leaf2->total;
  c1  = leaf1->cnts +(i = lvl->cnt);
  c2  = leaf2->cnts + i;        /* traverse the leaf counters */
  b   = lvl->buf    + i;        /* and the evaluation buffer */
  for (s = 0; --i >= 0; ) {     /* compute the evaluation */
    c = *--c1 + *--c2;          /* of a merged leaf, i.e., */
    t = (*--b < n) ? *b : n;    /* for summed counters */
    if ((c > 0) && (t > 0)) s += c *log(t / c);
  }                             /* update the mutual specificity */
  pt->res1[0] = pt->res0[0] +(s -leaf1->eval -leaf2->eval);
  pt->res1[1] = (double)pt->leafcnt-2; /* note normalization factor */
  return s;                     /* return the leaf evaluation */
}  /* mutspc() */

/*--------------------------------------------------------------------*/

static FTE fn_prob[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { (EVALFN*)0, MRG_INDEP  },
  /* PT_INFGAIN  1 */ { info,       MRG_INDEP  },
  /* PT_INFGBAL  2 */ { info,       MRG_LOCAL  },
  /* PT_INFGR    3 */ { info,       MRG_LOCAL  },
  /* PT_INFSGR1  4 */ { info,       MRG_LOCAL  },
  /* PT_INFSGR2  5 */ { info,       MRG_LOCAL  },
  /* PT_QIGAIN   6 */ { quad,       MRG_INDEP  },
  /* PT_QIGBAL   7 */ { quad,       MRG_LOCAL  },
  /* PT_QIGR     8 */ { quad,       MRG_LOCAL  },
  /* PT_QISGR1   9 */ { quad,       MRG_LOCAL  },
  /* PT_QISGR2  10 */ { quad,       MRG_LOCAL  },
  /* PT_GINI    11 */ { gini1,      MRG_INDEP  },
  /* PT_GINISYM 12 */ { gini2,      MRG_LOCAL  },
  /* PT_GINIMOD 13 */ { gini1,      MRG_LOCAL  },
  /* PT_RELIEF  14 */ { gini1,      MRG_LOCAL  },
  /* PT_WDIFF   15 */ { wdiff,      MRG_INDEP  },
  /* PT_CHI2    16 */ { chi2,       MRG_INDEP  },
  /* PT_CHI2NRM 17 */ { chi2,       MRG_LOCAL  },
  /* PT_WOEVID  18 */ { evid,       MRG_INDEP  },
  /* PT_RELEV   19 */ { relev,      MRG_INDEP  },
  /* PT_BDM     20 */ { bdm,        MRG_INDEP  },
  /* PT_BDMOD   21 */ { bdmod,      MRG_INDEP  },
  /* PT_RDLREL  22 */ { dlrel,      MRG_INDEP  },
  /* PT_RDLABS  23 */ { dlabs,      MRG_INDEP  },
  /* PT_STOCO   24 */ { stoco,      MRG_LOCAL  },
  /* PT_UNKNOWN 25 */ { (EVALFN*)0, MRG_INDEP  },
};                              /* functions for probability trees */

static FTE fn_poss[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { (EVALFN*)0, MRG_INDEP  },
  /* PT_SPCGAIN  1 */ { spec,       MRG_GLOBAL },
  /* PT_SPCGBAL  2 */ { spec,       MRG_GLOBAL },
  /* PT_SPCGR    3 */ { spec,       MRG_GLOBAL },
  /* PT_SPCSGR1  4 */ { spec,       MRG_GLOBAL },
  /* PT_SPCGR2   5 */ { spec,       MRG_GLOBAL },
  /*             6 */ { (EVALFN*)0, 0          },
  /*             7 */ { (EVALFN*)0, 0          },
  /*             8 */ { (EVALFN*)0, 0          },
  /*             9 */ { (EVALFN*)0, 0          },
  /*            10 */ { (EVALFN*)0, 0          },
  /*            11 */ { (EVALFN*)0, 0          },
  /*            12 */ { (EVALFN*)0, 0          },
  /*            13 */ { (EVALFN*)0, 0          },
  /*            14 */ { (EVALFN*)0, 0          },
  /* PT_WDIFF   15 */ { pdiff,      MRG_INDEP  },
  /* PT_CHI2    16 */ { pchi2,      MRG_INDEP  },
  /*            17 */ { (EVALFN*)0, 0          },
  /*            18 */ { (EVALFN*)0, 0          },
  /*            19 */ { (EVALFN*)0, 0          },
  /* PT_MUTSPC  20 */ { mutspc,     MRG_INDEP  },
  /*            21 */ { (EVALFN*)0, 0          },
  /*            22 */ { (EVALFN*)0, 0          },
  /*            23 */ { (EVALFN*)0, 0          },
  /*            24 */ { (EVALFN*)0, 0          },
  /*            25 */ { (EVALFN*)0, 0          },
};                              /* functions for possibility trees */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static PTLEAF** collect (PTLVL *lvl, PTNODE *node, PTLEAF **leaves)
{                               /* --- collect leaves in a subtree */
  VALID  i, k;                  /* loop variable */
  PTNODE *child;                /* to traverse the child nodes */

  assert(lvl && node && leaves);/* check the function arguments */
  if (lvl->index < 0) {         /* if at the leaf level */
    if (((PTLEAF*)node)->total <= 0)
      return leaves;            /* check whether leaf is mergable */
    *leaves = (PTLEAF*)node;    /* add the leaf to the vector */
    return ++leaves;            /* and return the new end */
  }                             /* of the leaf vector */
  for (k = (lvl++)->cnt, i = 0; i < k; i++) {
    child = node->children[i];  /* traverse the children */
    if (child && (child->parent == node) && (child->index == i))
      leaves = collect(lvl, child, leaves);
  }                             /* collect the leaves recursively and */
  return leaves;                /* return the new end of the vector */
}  /* collect() */

/*--------------------------------------------------------------------*/

static double local (PTREE *pt, MRGTAB *tab)
{                               /* --- do local structure learning */
  VALID  i, k;                  /* loop variables, leaf counter */
  PTLEAF *m1, *m2;              /* leaves to be merged */
  double curr, best, t;         /* buffers for merger evaluation */
  WEIGHT *s, *d;                /* to traverse the counters */
  VALID  r1, r2 = 0;            /* merge table row indices */

  assert(pt && tab);            /* check the function arguments */

  /* --- initialize the merge table --- */
  curr = pt->eval;              /* get the current evaluation */
  if (tab->evals) {             /* if merge gains are independent */
    for (i = tab->rowcnt; --i > 0; ) {
      best = -INFINITY;         /* traverse the table rows */
      for (k = i; --k >= 0; ) { /* traverse the preceding rows */
        tab->evalfn(pt, tab->leaves[i], tab->leaves[k]);
        tab->evals[i][k] = t = pt_reeval(pt) -curr;
        if (t >= best) { best = t; tab->bids[i] = k; }
      }                         /* note the gain that results from */
    }                           /* a merger and find the best merger */
  }                             /* in each table row */

  while (1) {                   /* leaf merge loop */

    /* --- evaluate the mergers --- */
    best = -INFINITY; r1 = -1;  /* initialize the best evaluation */
    if (!tab->evals) {          /* -- if there is no evaluation part */
      for (i = tab->rowcnt; --i > 0; ) {
        m1 = tab->leaves[i];    /* traverse the mergable leaves */
        if (!m1) continue;      /* (not yet merged to another) */
        for (k = i; --k >= 0; ) {
          m2 = tab->leaves[k];  /* traverse the remaining leaves and */
          if (!m2) continue;    /* evaluate the possible mergers */
          tab->evalfn(pt, m1, m2);
          t = pt_reeval(pt) -curr;
          if (t >= best) { best = t; r1 = k; r2 = i; }
        }                       /* determine the pair of leaves */
      } }                       /* that yield the best evaluation */
    else {                      /* -- if there is an evaluation part */
      for (i = tab->rowcnt; --i > 0; ) {
        k = tab->bids[i];       /* traverse the rows of the table */
        if (k < 0) continue;    /* that contain valid mergers and */
        t = tab->evals[i][k];   /* get the best merger in the row */
        if (t >= best) { best = t; r1 = k; r2 = i; }
      }                         /* find the best leaf merger */
    }                           /* in the whole table */
    if ((r1   < 0)              /* if there is no leaf merger */
    ||  (best < tab->minimp))   /* or the best merger is not */
      break;                    /* good enough, abort the loop */
    assert(r1 < r2);            /* check the leaf indices */

    /* --- execute the best merger --- */
    m1 = tab->leaves[r1];       /* get the two leaves to be merged, */
    m2 = tab->leaves[r2];       /* check whether they are valid, and */
    assert(m1 && m2);           /* compute the new evaluation term */
    m1->eval = tab->evalfn(pt, m1, m2);
    curr     = pt_reeval(pt);   /* update the current evaluation */
    for (i = 5; --i >= 0; )     /* and copy the evaluation buffers */
      pt->res0[i] = pt->res1[i];/* from which it is computed */
    m1->pathcnt += m2->pathcnt; /* sum the paths to the leaves */
    s = m2->cnts +(i = pt->levels[pt->concnt].cnt);
    d = m1->cnts + i;           /* merge the counters of the leaves */
    if (pt->type & PT_POSS) {   /* in poss. mode by taking maximum */
      while (--i >= 0) { if (*--s > *--d) *d = *s; }
      if (m2->total > m1->total) m1->total = m2->total; }
    else {                      /* in prob. mode merge by summing */
      while (--i >= 0) *--d += *--s;
      m1->total += m2->total;   /* merge also the leaf totals */
    }
    m2->total   = 0;            /* mark the source leaf as merged */
    m2->pathcnt = 0;            /* (it cannot be deleted immediately) */
    m2->parent  = (PTNODE*)m1;  /* and note the destination leaf */
    if (--tab->leafcnt <= 2)    /* decrement the leaf counter and */
      break;                    /* abort if there are only two leaves */

    /* --- update the merge table --- */
    tab->leaves[r2] = NULL;     /* invalidate the second leaf */
    if (!tab->evals) continue;  /* check for a leaf merge table */
    tab->bids[r2] = -1;         /* invalidate the indices of */
    tab->bids[r1] = -1;         /* the best leaf mergers */
    for (best = -INFINITY, k = r1; --k >= 0; ) {
      if (!tab->leaves[k])      /* traverse the unmerged leaves */
        continue;               /* in the first leaf's row */
      tab->evalfn(pt, m1, tab->leaves[k]);
      tab->evals[r1][k] = t = pt_reeval(pt) -curr;
      if (t >= best) { best = t; tab->bids[r1] = k; }
    }                           /* update the merger evaluations */
    for (i = tab->rowcnt; --i > r1; ) {
      if (!tab->leaves[i])      /* traverse the following rows */
        continue;               /* that contain unmerged leaves */
      tab->evalfn(pt, m1, tab->leaves[i]);
      tab->evals[i][r1] = t = pt_reeval(pt) -curr;
      k = tab->bids[i];         /* reevaluate mergers with first leaf */
      if ((k != r1) && (k != r2)        /* check whether a search */
      &&  (t > tab->evals[i][k])) {     /* for the index of the */
        tab->bids[i] = r1; continue; }  /* best merger is necessary */
      for (best = -INFINITY, k = i; --k >= 0; ) {
        if (!tab->leaves[k]) continue;
        t = tab->evals[i][k];   /* traverse the unmerged leaves */
        if (t >= best) { best = t; tab->bids[i] = k; }
      }                         /* find the best merger among */
    }                           /* those represented in this row */

  }  /* while (1) */

  return pt->eval = curr;       /* return the current evaluation */
}  /* local() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

double pt_local (PTREE *pt, int lwise, double minimp)
{                               /* --- do local structure learning */
  VALID  i, k, n = 0;           /* loop variables, leaf counter */
  ATTID  c;                     /* number of conditions/levels */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf, *dest, **p;     /* to traverse the leaves */
  PTNODE *node;                 /* to traverse the nodes one level up */
  double *e;                    /* to organize the merge table */
  FTE    *fte;                  /* function table element */
  MRGTAB *tab;                  /* leaf merge table */

  assert(pt);                   /* check the function arguments */
  if ((pt->concnt <= 0)         /* if there are no conditions */
  ||  (pt->total  <= 0))        /* or all counters are zero, */
    return pt->eval;            /* return the existing evaluation */

  /* --- create a merge table --- */
  lvl = pt->levels +pt->concnt; /* traverse the leaf level */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    if (leaf->total > 0) n++;   /* count the mergable leaves */
  if (n <= 2) return pt->eval;  /* there must be more than two leaves */
  tab = (MRGTAB*)malloc(sizeof(MRGTAB) +(size_t)(n-1) *sizeof(PTLEAF*));
  if (!tab) return NAN;         /* create a leaf merge table */
  tab->leafcnt = n;             /* and initialize it */
  tab->minimp  = minimp;
  fte = ((pt->type == PT_PROB) ? fn_prob : fn_poss) +pt->measure;
  if (!fte->evalfn) return NAN; /* get the evaluation function and */
  tab->evalfn = fte->evalfn;    /* store it in the leaf merge table */
  if (fte->type != MRG_INDEP)   /* if merge gains are dependent */
    tab->evals = NULL;          /* do not create table elements */
  else {                        /* if merge gains are independent */
    tab->bids  = (VALID*)malloc((size_t)n *sizeof(VALID));
    if (!tab->bids) { free(tab); }
    tab->evals = (double**)malloc((size_t)n *sizeof(double*));
    if (!tab->evals) { free(tab->bids); free(tab); }
    e = (double*)malloc(((size_t)n *(size_t)(n-1)) /2 *sizeof(double));
    if (!e) { free(tab->evals); free(tab->bids); free(tab); }
    for (i = 0; i < n; i++) { tab->evals[i] = e; e += i; }
  }                             /* create evaluation part of table */

  /* --- do local structure learning --- */
  c   = (lwise > 0) ? pt->concnt +1 : 1;
  lvl = pt->levels +c;          /* get the number of levels and */
  while (--c >= 0) {            /* traverse the tree levels */
    for (node = (--lvl)->list; node; node = node->succ) {
      tab->rowcnt = (VALID)(collect(lvl, lvl->list, tab->leaves)
                                                   -tab->leaves);
      local(pt, tab);           /* collect all mergable leaves */
    }                           /* and do local structure learning */
  }                             /* on each tree level */

  /* --- delete the merge table --- */
  if (tab->evals) {             /* delete the evaluation entries */
    free(tab->bids); free(tab->evals[0]); free(tab->evals); }
  free(tab);                    /* delete the leaf merge table */

  /* --- finalize the mergers --- */
  lvl = pt->levels +pt->concnt; /* traverse the leaf level */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->pathcnt > 0) continue; /* traverse the merged leaves */
    dest = (PTLEAF*)leaf->parent;    /* follow references to dests. */
    while (dest->pathcnt <= 0) dest = (PTLEAF*)dest->parent;
    leaf->parent = (PTNODE*)dest;    /* find the final destination */
  }                                  /* and set the reference to it */
  k = (lvl-1)->cnt;             /* traverse level above the leaves */
  for (node = (lvl-1)->list; node; node = node->succ) {
    for (p = (PTLEAF**)node->children +(i = k); --i >= 0; ) {
      if (!*--p) continue;      /* traverse the children of each node */
      if ((*p)->pathcnt <= 0) *p = (PTLEAF*)(*p)->parent;
    }                           /* if a child has been merged, */
  }                             /* set the link to the destination */
  for (p = (PTLEAF**)&lvl->list; *p; ) {
    leaf = *p;                  /* traverse the leaf nodes again */
    if (leaf->pathcnt > 0)      /* if a leaf has not been merged, */
      p = (PTLEAF**)&leaf->succ;/* just go to the next leaf */
    else {                      /* if a leaf has been merged */
      *p = (PTLEAF*)leaf->succ; free(leaf); }
  }                             /* delete the leaf */
  return pt_eval(pt, pt->measure, pt->params);
}  /* pt_local() */             /* return the new tree evaluation */
